package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.*;
import ru.dbotthepony.mc.otm.capability.matter.*;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.menu.MatterPanelMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.util.*;
import java.util.stream.Collectors;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityMatterPanel extends BlockEntityMattery implements IMatterGridCell, IMatterTaskProvider {
	public BlockEntityMatterPanel(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_PANEL, p_155229_, p_155230_);
	}

	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.matter_panel");

	private final ArrayList<MatterPanelMenu> listeners = new ArrayList<>();

	public void attachMenu(MatterPanelMenu menu) {
		listeners.add(menu);
	}

	public void deatachMenu(MatterPanelMenu menu) {
		listeners.remove(menu);
	}

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterPanelMenu(containerID, inventory, this);
	}

	private boolean valid = true;

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
		resolver.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
		resolver = LazyOptional.of(() -> this);
	}

	private LazyOptional<BlockEntityMatterPanel> resolver = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid && (cap == MatteryCapability.MATTER_CELL || cap == MatteryCapability.TASK))
			return resolver.cast();

		return super.getCapability(cap, side);
	}

	private MatterGrid grid;

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			scheduleDiscoverNeighbours(p_155231_, getBlockPos());
	}

	@Override
	public void setRemoved() {
		super.setRemoved();

		if (level != null && !level.isClientSide && grid != null)
			grid.remove(this);
	}

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}

	@Nullable
	@Override
	public IMatterTaskProvider getTaskProvider() {
		return this;
	}

	private final HashMap<UUID, MatterTask> tasks = new HashMap<>();

	@Nonnull
	@Override
	public Collection<MatterTask> getTasks() {
		return tasks.values().stream().filter(task -> task.required() > 0).collect(Collectors.toList());
	}

	@Nonnull
	@Override
	public Collection<MatterTask> getAllTasks() {
		return List.copyOf(tasks.values());
	}

	@Nullable
	@Override
	public MatterTaskAllocation allocateTask(boolean simulate) {
		if (grid == null)
			return null;

		for (var entry : tasks.entrySet()) {
			var task = entry.getValue();

			if (task.required() > 0) {
				var get_pattern = grid.getPattern(task.pattern());

				if (get_pattern != null) {
					if (!simulate) {
						var newer = task.shrinkRequired(1);
						tasks.put(entry.getKey(), newer);
						listeners.forEach(menu -> menu.taskUpdated(newer));
						grid.onMatterTaskUpdated(newer, task);
						setChanged();
					}

					return new MatterTaskAllocation(task, get_pattern);
				}
			}
		}

		return null;
	}

	@Override
	public boolean notifyTaskCompletion(MatterTask task) {
		var get_task = tasks.get(task.id());

		if (get_task == null) {
			return false;
		}

		var old_task = get_task;
		get_task = get_task.shrinkInProgress(1);

		if (get_task.required() <= 0 && get_task.in_progress() <= 0) {
			tasks.remove(task.id());

			if (grid != null)
				grid.onMatterTaskCreated(task);

			MatterTask finalGet_task1 = get_task;
			listeners.forEach(menu -> menu.taskRemoved(finalGet_task1));
		} else {
			tasks.put(task.id(), get_task);

			if (grid != null) {
				grid.onMatterTaskUpdated(get_task, old_task);
			}

			MatterTask finalGet_task = get_task;
			listeners.forEach(menu -> menu.taskUpdated(finalGet_task));
		}

		setChanged();

		return true;
	}

	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		ListTag list = new ListTag();

		for (MatterTask task : tasks.values()) {
			list.add(task.serializeNBT());
		}

		nbt.put("tasks", list);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);
		tasks.clear();

		ListTag list = nbt.getList("tasks", Tag.TAG_COMPOUND);

		for (Tag tag : list) {
			MatterTask task = MatterTask.deserializeNBT(tag);

			if (task != null) {
				tasks.put(task.id(), task);
			}
		}
	}

	@Nullable
	@Override
	public MatterTask getTask(UUID id) {
		return tasks.get(id);
	}

	public void removeTask(UUID id) {
		var task = tasks.get(id);

		if (task == null)
			return;

		tasks.remove(id);

		if (grid != null)
			grid.onMatterTaskRemoved(task);

		listeners.forEach(menu -> menu.taskRemoved(task));
		setChanged();
	}

	public void removeTask(PatternState state) {
		removeTask(state.id());
	}

	public MatterTask addTask(PatternState state, int how_much) {
		var task = new MatterTask(UUID.randomUUID(), state.id(), state.item(), 0, 0, how_much);
		tasks.put(task.id(), task);

		if (grid != null)
			grid.onMatterTaskCreated(task);

		listeners.forEach(menu -> menu.taskUpdated(task));

		setChanged();

		return task;
	}

	@Override
	public void dropAllTasks() {
		tasks.clear();
	}
}
